// client.cpp -- Wedges Internet UNO Client
//
// By Andrew Wedgbury
// Copyright 1997

#define STRICT
#include <windows.h>
#include <winsock.h>
#include <windowsx.h>
#include <commctrl.h>
#include <string.h>
#include <stdio.h>
#include <time.h>
#pragma hdrstop

#include "resource.h"
#include "../NetUNO.h"

//====================================================== GLOBAL DATA ====
const char	szAppName[] = "Wedge's NetUNO";
const char	szVersion[] = "Version 1.00 (32bit)";
const char	szMainClassName[] = "WUNOClient";

#define MAX_PENDING_CONNECTS	4 
#define NO_FLAGS_SET			0
#define UNO_MSG_LENGTH			400
#define WSA_ACCEPT			(WM_USER + 0)
#define WSA_READ			(WM_USER + 1)
#define WSA_LOOKUP			(WM_USER + 2)

#define	CARD_OVERLAP			12
#define	HANDX					100
#define HANDY					110

// MouseOver flags
#define OVER_NOTHING	0
#define	OVER_PICKUP		1
#define	OVER_PUTDOWN	2
#define	OVER_HAND		3
#define	OVER_PALETTE	4

HINSTANCE	hInst;						// Windows data
HWND		hMainWin,hChatDlg,hDealerDlg;
UINT		nInstances=0;

HBRUSH		hBkgBrush;					// GDI objects
HPEN		hYellowPen,hBlackPen;

SOCKET		sock;						// WinSock data
u_short		portno = 1527;
int			status;
SOCKADDR_IN dest_sin;

char		szBuff[UNO_MSG_LENGTH];		// Text buffers
char		szSend[UNO_MSG_LENGTH+30];
char		szParse[200];
char		szDispText[200];
char		szUserName[40];
char		szServerName[80];

HAND		*MyHand,*MyPal;				// Cards
CARD		PileCards[10];
UINT		nPileCards=0;
UINT		nMouseOverCard=0;

GAMESTATS	GameStats;					// Game info
WORD		nPlayers;
UINT		nOurIndex=0;
char		szPlayersNames[10][40];
UINT		nPlayersCards[10];
BOOL		bGameRunning=FALSE;

UINT		bDirFwds=TRUE;
UINT		nTurn=0;
UINT		bOurTurn=FALSE;
CARD		cTopCard=0;
UINT		nPickup=0;
UINT		nAceSuit=0;

//======================== Function Prototypes ==
int PASCAL			WinMain(HINSTANCE,HINSTANCE,LPSTR,int);

LRESULT CALLBACK	MainWndProc(HWND,UINT,WPARAM,LPARAM);	// WNDPROC's
LRESULT CALLBACK	SubChatTypeWndProc(HWND,UINT,WPARAM,LPARAM);

BOOL PASCAL			ChatDlgProc(HWND,UINT,WPARAM,LPARAM);
BOOL PASCAL			ConnectDlgProc(HWND,UINT,WPARAM,LPARAM);
BOOL PASCAL			AboutDlgProc(HWND,UINT,WPARAM,LPARAM);
BOOL PASCAL			DealerDlgProc(HWND,UINT,WPARAM,LPARAM);

void				DoSocketMsg(SOCKET,WORD);
BOOL				ProgramInit(void);





UINT Parse(char szSrc[],char *szDest)
{
	UINT i;
	for (i=0; i<200; i++)
	{
		if (szSrc[i] == '$') break;
		*(szDest + i) = szSrc[i];	
	}
	*(szDest + i) = 0;
	return i+1;
}





void ExecuteSocketMsg(char szMsg[])
{
	UINT x,n,d;
	switch (szMsg[4])				// SERVER -> CLIENT Messages
	{
		case 'C': // CONNECT CONFIRMED
			n = 5;
			n += Parse(&szMsg[n],szParse); nOurIndex = atoi(szParse);
			sprintf(szSend,"NetUNO on %s",szServerName);
			SetWindowText(hMainWin,szSend);
			sprintf(szSend,"Chat window - %s",szUserName);
			SetWindowText(hChatDlg,szSend);
			break;

		case 'T': // TEXT CHAT TO DISPLAY
			n = 5;
			n += Parse(&szMsg[n],szParse);
			memcpy(szSend,&szMsg[n],atoi(szParse)); szSend[atoi(szParse)] =0;
			lstrcat(szSend,"\x0D\x0A"); // 0x0D 0x0A
			d = GetWindowTextLength(GetDlgItem(hChatDlg,IDCE_CHAT_DISP));		
			SendDlgItemMessage(hChatDlg,IDCE_CHAT_DISP,EM_SETSEL,(WPARAM)d,(LPARAM)d);
			SendDlgItemMessage(hChatDlg,IDCE_CHAT_DISP,EM_REPLACESEL,(WPARAM)FALSE,(LPARAM)szSend);
			break;

		case 'N': // NEW TURN
			n = 5;
			n += Parse(&szMsg[n],szParse);	bDirFwds = atoi(szParse);
			n += Parse(&szMsg[n],szParse);	nTurn = atoi(szParse);
			n += Parse(&szMsg[n],szParse);	bOurTurn = atoi(szParse);
			n += Parse(&szMsg[n],szParse);	nPickup = atoi(szParse);
			n += Parse(&szMsg[n],szParse);	nAceSuit = atoi(szParse);
			n += Parse(&szMsg[n],szParse);	nPileCards = atoi(szParse);

			for (x=0; x<nPileCards; x++)
			{
				n += Parse(&szMsg[n],szParse); PileCards[x] = atoi(szParse);
			}
			cTopCard = PileCards[nPileCards-1];

			for (x=0; x<nPlayers; x++)
			{
				n += Parse(&szMsg[n],szParse); nPlayersCards[x] = atoi(szParse);
			}
			if (bOurTurn) SetForegroundWindow(hMainWin); // Bring us to the top on our turn
			if (nPickup) MessageBeep(MB_ICONEXCLAMATION); // Alert on pickup
			InvalidateRect(hMainWin,NULL,FALSE);
			break;

		case 'U': // PICKED UP CARD
			n = 5;
			n += Parse(&szMsg[n],szParse); 
			MyHand->GiveCard(atoi(szParse));
			InvalidateRect(hMainWin,NULL,FALSE);
			break;

		case 'W': // WINNER
			break;

		case 'G': // NEW GAME
			delete MyHand; MyHand = new HAND(FALSE);
			delete MyPal; MyPal = new HAND(FALSE);
			bGameRunning=TRUE;
			InvalidateRect(hMainWin,NULL,TRUE);
			break;

		case 'D': // GAME DATA UPDATE
			while (SendDlgItemMessage(hDealerDlg,IDCL_NAMES,LB_GETCOUNT,0,0))
			{
				SendDlgItemMessage(hDealerDlg,IDCL_NAMES,LB_DELETESTRING,0,0);
				SendDlgItemMessage(hDealerDlg,IDCC_START_WITH,CB_DELETESTRING,0,0);
			}
			n = 5;
			n += Parse(&szMsg[n],szParse); nPlayers = atoi(szParse);
			for (x=0; x<nPlayers; x++)
			{
				n += Parse(&szMsg[n],szParse); lstrcpy(szPlayersNames[x],szParse);
				SendDlgItemMessage(hDealerDlg,IDCL_NAMES,LB_ADDSTRING,0,(LPARAM)(LPCTSTR)szPlayersNames[x]);
				SendDlgItemMessage(hDealerDlg,IDCC_START_WITH,CB_ADDSTRING,0,(LPARAM)(LPCTSTR)szPlayersNames[x]);
			}
			SendDlgItemMessage(hDealerDlg,IDCC_START_WITH,CB_SETCURSEL,0,0);
			break;

		case 'S':
			if (szMsg[5] == '1')
				ShowWindow(hDealerDlg,SW_SHOWNORMAL);
			else
				ShowWindow(hDealerDlg,SW_HIDE);
			UpdateWindow(hDealerDlg);
			break;
	}
}


//========================================== Handle socket messages =====
void DoSocketMsg(SOCKET s,WORD msg)
{
	UINT i;
	if (msg == FD_READ) // Socket READ
	{
		status = recv(s,szBuff,UNO_MSG_LENGTH,NO_FLAGS_SET);
		if (status) 
		{
			szBuff[status] = '\0';
			for (i=0; i<(UINT)status; i++)
			{
				if (( (szBuff[i] == 'W') && (szBuff[i+1] == 'N') ) && ( (szBuff[i+2] == 'U') && (szBuff[i+3] == 'S') ))
				{
					ExecuteSocketMsg(&szBuff[i]);
				}
			}
			return;
		}
	}

	// Socket CLOSE
}



//======================================================= Chat system ============

// SubClass info for chat type edit control 
WNDPROC ChatTypeWndProc;

LRESULT CALLBACK SubChatTypeWndProc(HWND hWnd,UINT message,WPARAM wParam,LPARAM lParam)
{
	if ((message == WM_CHAR) && ((int)wParam == 13)) // Intercept returns 
	{
		GetDlgItemText(hChatDlg,IDCE_CHAT_TYPE,szBuff,sizeof(szBuff));
		sprintf(szSend,"WNUCT%u$%s",lstrlen(szBuff),szBuff);
		send(sock,szSend,strlen(szSend),NO_FLAGS_SET);
		SetDlgItemText(hChatDlg,IDCE_CHAT_TYPE,"");
		return 0;
	}
	return CallWindowProc(ChatTypeWndProc,hWnd,message,wParam,lParam);
}


BOOL PASCAL ChatDlgProc(HWND hDlg,UINT message,WPARAM wParam,LPARAM lParam)
{
	switch(message)
	{
		case WM_INITDIALOG:
			ChatTypeWndProc = (WNDPROC)GetWindowLong(GetDlgItem(hDlg,IDCE_CHAT_TYPE),GWL_WNDPROC);
			SetWindowLong(GetDlgItem(hDlg,IDCE_CHAT_TYPE),GWL_WNDPROC,(LONG)SubChatTypeWndProc);
			return(TRUE);

		case WM_COMMAND:
			break;
	}
	return(FALSE);
}







BOOL PASCAL DealerDlgProc(HWND hDlg,UINT message,WPARAM wParam,LPARAM lParam)
{
	UINT x;
	switch(message)
	{
		case WM_INITDIALOG:
			SetDlgItemText(hDlg,IDCE_NUM_DECKS,"1");
			SetDlgItemText(hDlg,IDCE_NUM_CARDS,"7");
			CheckRadioButton(hDlg,IDCR_DIR_FW,IDCR_DIR_BK,IDCR_DIR_FW);
			CheckRadioButton(hDlg,IDCR_RUNS_Y,IDCR_RUNS_N,IDCR_RUNS_Y);
			SendDlgItemMessage(hDlg,IDCC_START_WITH,CB_SETCURSEL,0,0);
			SendDlgItemMessage(hDlg,IDCL_NAMES,LB_SETCURSEL,0,0);
			return(TRUE);

		case WM_COMMAND:
			switch (LOWORD(wParam))
			{
				case IDCB_NEW_GAME:
					sprintf(szSend,"WNUCG%u$%u$%u$%u$%u$",
						GetDlgItemInt(hDlg,IDCE_NUM_DECKS,NULL,FALSE),
						GetDlgItemInt(hDlg,IDCE_NUM_CARDS,NULL,FALSE),
						SendDlgItemMessage(hDlg,IDCC_START_WITH,CB_GETCURSEL,0,0),
						IsDlgButtonChecked(hDlg,IDCR_DIR_FW),
						IsDlgButtonChecked(hDlg,IDCR_RUNS_Y));
					send(sock,szSend,strlen(szSend),NO_FLAGS_SET);
					break;

				case IDCB_SET_DEALER:
					x = SendDlgItemMessage(hDlg,IDCL_NAMES,LB_GETCURSEL,0,0);
					if ((x<nPlayers) && (x!=nOurIndex))
					{
						sprintf(szSend,"WNUCS%u$",x);
						send(sock,szSend,strlen(szSend),NO_FLAGS_SET);
					}
					break;

				case IDCB_KICK_OUT:
					break;
			}
			break;
	}
	return(FALSE);
}




BOOL ConnectToServer(void)
{
	// ================ Connect to server ================
	sock = socket(AF_INET,SOCK_STREAM,0);
	if (sock == INVALID_SOCKET) 
	{
		MessageBox(hMainWin,"socket() failed","Error",MB_OK);
		return FALSE;
	}
	dest_sin.sin_family = AF_INET;

	if (connect(sock,(PSOCKADDR)&dest_sin,sizeof(dest_sin)) < 0) 
	{
		closesocket(sock);
		MessageBox(hMainWin,"Connection failed","Error",MB_OK);
		return FALSE;
	}

	if ((status = WSAAsyncSelect(sock,hMainWin,WSA_READ,FD_READ | FD_CLOSE )) > 0) 
	{
		wsprintf(szBuff,"%d (0x%x)");
		MessageBox(hMainWin,"Error on WSAAsyncSelect()",szBuff,MB_OK);
		closesocket(sock);
	}

	// Send the server our name
	sprintf(szSend,"WNUCC%s",szUserName);
	send(sock,szSend,strlen(szSend),NO_FLAGS_SET);

	EnableMenuItem(GetMenu(hMainWin),IDM_CONNECT,MF_GRAYED);
	EnableMenuItem(GetMenu(hMainWin),IDM_DISCONNECT,MF_ENABLED);
	return TRUE;
}




//======================================================= Connect dialog ============
BOOL PASCAL ConnectDlgProc(HWND hDlg,UINT message,WPARAM wParam,LPARAM lParam)
{
	static PHOSTENT phe;
	static HANDLE hGetHost;
	switch(message)
	{
		case WM_INITDIALOG:
			gethostname(szBuff,sizeof(szBuff));
			SetDlgItemText(hDlg,IDCE_CON_SERVER,szBuff);
			SetDlgItemText(hDlg,IDCE_CON_PORT,"1527");
			SetFocus(GetDlgItem(hDlg,IDCE_CON_NAME)); 
			phe = (PHOSTENT)GlobalAlloc(GMEM_FIXED,MAXGETHOSTSTRUCT);
			return(FALSE);

		case WM_COMMAND:
			switch (LOWORD(wParam))
			{
				case IDCB_CONNECT:
					{
						if (hGetHost)
							WSACancelAsyncRequest(hGetHost);

						// ================ Validate dialog entries ================
						if (!GetDlgItemText(hDlg,IDCE_CON_NAME,szUserName,sizeof(szUserName)))
						{
							MessageBox(hMainWin,"Please enter a name","Error",MB_OK);
							return TRUE;
						}

						BOOL bSucess=FALSE;
						portno = (u_short)GetDlgItemInt(hDlg,IDCE_CON_PORT,&bSucess,FALSE);
						if (!bSucess) 
						{
							MessageBox(hMainWin,"Invalid port no.","Error",MB_OK);
							return TRUE;
						}
						dest_sin.sin_port = htons(portno);

						if (!GetDlgItemText(hDlg,IDCE_CON_SERVER,szServerName,sizeof(szServerName)))
						{
							MessageBox(hMainWin,"Please enter a server","Error",MB_OK);
							return TRUE;
						}
						
						unsigned long addr = inet_addr(szServerName);

						if (addr==INADDR_NONE)
							hGetHost = WSAAsyncGetHostByName(hDlg,WSA_LOOKUP,szServerName,(LPSTR)phe,MAXGETHOSTSTRUCT);
						else
						{
							memcpy((char FAR *)&(dest_sin.sin_addr),(LPSTR)&addr,4);
							if (ConnectToServer())
								EndDialog(hDlg,1);
							return TRUE;
						}

						if (hGetHost == NULL) 
						{
							MessageBox(hMainWin,"Could not start server lookup","Error",MB_OK);
							return TRUE;
						}
						SetCursor(LoadCursor(NULL,IDC_WAIT));
					}
					return TRUE;

				case IDCANCEL:
					if (hGetHost)
						WSACancelAsyncRequest(hGetHost);
					SetCursor(LoadCursor(NULL,IDC_ARROW));
					EndDialog(hDlg,0);
					return TRUE;
			}
			break;

		case WSA_LOOKUP:
			SetCursor(LoadCursor(NULL,IDC_ARROW));
			hGetHost=NULL;
			if ((WSAGETASYNCERROR(lParam)) || (!phe->h_length))
			{
				MessageBox(hMainWin,"Can't find server","Error",MB_OK);
				return TRUE;
			}
			memcpy((char FAR *)&(dest_sin.sin_addr),phe->h_addr,phe->h_length);
			if (ConnectToServer())
				EndDialog(hDlg,1);
			break;

		case WM_DESTROY:
			GlobalFree((HGLOBAL)phe);
			break;
	}
	return(FALSE);
}





//======================================================= About dialog ============
BOOL PASCAL AboutDlgProc(HWND hDlg,UINT message,WPARAM wParam,LPARAM lParam)
{
	switch(message)	// Also handles suit select dialog
	{
		case WM_INITDIALOG:
			SetDlgItemText(hDlg,IDCS_VERSION,szVersion);
			SetDlgItemText(hDlg,IDCE_EMAIL,"aw@wedger.demon.co.uk");
			SetDlgItemText(hDlg,IDCE_WWW,"http://www.wedger.demon.co.uk/uno.html");
			return(FALSE);

		case WM_COMMAND:
			switch (LOWORD(wParam))
			{
				case IDOK:
					EndDialog(hDlg,0);
					break;
				case 300:
				case 301:
				case 302:
				case 303:
					EndDialog(hDlg,LOWORD(wParam)-300);
					break;
			}
			return (TRUE);

		case WM_DRAWITEM:
			{
				LPDRAWITEMSTRUCT lpDrawStruct = (LPDRAWITEMSTRUCT)lParam;
				HDC hDC = lpDrawStruct->hDC;
				RECT rWin;
				GetClientRect(GetDlgItem(hDlg,wParam),&rWin);
				HDC hBMPDC = CreateCompatibleDC(hDC);
				HBITMAP hBMP = LoadBitmap(hInst,(LPSTR)(UINT)wParam);
				HGDIOBJ hSaveObj = SelectObject(hBMPDC,hBMP);
				BitBlt(hDC,0,0,rWin.right,rWin.bottom,hBMPDC,0,0,SRCCOPY);
				SelectObject(hBMPDC,hSaveObj);
				DeleteDC(hBMPDC);
				DeleteObject(hBMP);
			}
			return(TRUE);
	}
	return(FALSE);
}





UINT CheckMouseOver(UINT x,UINT y)
{
	RECT rWin; GetClientRect(hMainWin,&rWin);
	UINT xc = 100; //rWin.right/2;
	UINT yc = rWin.bottom/2;
	nMouseOverCard=0;

	// Check Pick-up deck or put down pile16x24
	if ( (y>70) && (y<94) )
	{
		if ( (x>(xc-20)) && (x<(xc-4)) ) return OVER_PICKUP;
		if ( (x>(xc+6)) && (x<(xc+6+(CARD_OVERLAP*nPileCards))) ) return OVER_PUTDOWN;
	}

	// Check own hand
	if ( ((y>(HANDY+30)) && (y<(HANDY+54))) && (x>HANDX) )
	{
		if (x<(HANDX+(CARD_OVERLAP*MyHand->GetNumCards())))
		{
			nMouseOverCard = ((x-HANDX) / CARD_OVERLAP);
			return OVER_HAND;	
		}
	}

	// Check pick-up / put-down area
	if ( ((y>HANDY) && (y<HANDY+24)) && (x>HANDX) )
	{
		if (x<(HANDX+(CARD_OVERLAP*MyPal->GetNumCards())))
		{
			nMouseOverCard = ((x-HANDX) / CARD_OVERLAP);
			return OVER_PALETTE;
		}
	}

	return OVER_NOTHING;
}






void PaintCard(HDC hDC,CARD c,UINT x,UINT y)
{
	HBITMAP hBMP = LoadBitmap(hInst,(LPCTSTR)100+c);
	HDC hBMPDC = CreateCompatibleDC(hDC);
	HGDIOBJ hSaveObj = SelectObject(hBMPDC,hBMP);
	BitBlt(hDC,x,y,16,24,hBMPDC,0,0,SRCCOPY);
	SelectObject(hBMPDC,hSaveObj);
	DeleteObject(hBMP);
	DeleteDC(hBMPDC);
}


void PaintOpponent(HDC hDC,UINT ix,UINT i)
{
	UINT nPOff=100;
	SetTextColor(hDC,RGB(0,0,0));
	ExtTextOut(hDC,21+(nPOff*ix),11,NULL,NULL,szPlayersNames[i],strlen(szPlayersNames[i]),(LPINT)NULL);
	if (nTurn  == i)
		SetTextColor(hDC,RGB(255,255,0));
	else
		SetTextColor(hDC,RGB(255,255,255));
	SelectObject(hDC,GetStockObject(SYSTEM_FONT));
	ExtTextOut(hDC,20+(nPOff*ix),10,NULL,NULL,szPlayersNames[i],strlen(szPlayersNames[i]),(LPINT)NULL);

	for (UINT x=0; x<nPlayersCards[i]; x++)
	{
		PaintCard(hDC,100,20+(nPOff*ix)+(4*x),30);
	}
}


//===================================================
void PaintMainWindow(HWND hWnd)
{
//	HGDIOBJ	hSaveObj;
	PAINTSTRUCT ps;
	UINT i,x=0;
	static UINT nPrevNumCardsHand;
	static UINT nPrevNumCardsPal;
	
	HDC hDC = BeginPaint(hWnd,&ps);				//--PAINT START---

	SetBkMode(hDC,TRANSPARENT);
	SelectObject(hDC,GetStockObject(SYSTEM_FONT));
	SelectObject(hDC,hBkgBrush);
	SelectObject(hDC,GetStockObject(NULL_PEN));
	RECT rWin; GetClientRect(hMainWin,&rWin);

	// Paint the opponents along the top
	Rectangle(hDC,20,29,rWin.right,60);
	for (i=(nOurIndex+1); i<nPlayers; i++)
		PaintOpponent(hDC,x++,i);
	for (i=0; i<nOurIndex; i++)
		PaintOpponent(hDC,x++,i);

	if (bGameRunning)
	{

		// Paint the deck and pile in the middle
		PaintCard(hDC,100,80,70);
		Rectangle(hDC,106,69,120+(5*CARD_OVERLAP),95);
		for (i=0; i<nPileCards; i++)
			PaintCard(hDC,PileCards[i],106+(i*CARD_OVERLAP),70);
		
		// Paint status text on left
		Rectangle(hDC,10,HANDY-1,90,rWin.bottom);
		SetTextColor(hDC,RGB(255,255,0));
		if (bOurTurn)
		{
			lstrcpy(szSend,"Your turn");
			ExtTextOut(hDC,10,HANDY,NULL,NULL,szSend,strlen(szSend),(LPINT)NULL);
			if (nPickup)
			{
				sprintf(szSend,"Pickup %u",nPickup);
				ExtTextOut(hDC,10,HANDY+20,NULL,NULL,szSend,strlen(szSend),(LPINT)NULL);
			}
		}
		if (CNUM(cTopCard)==1)
		{
			switch(nAceSuit)
			{
				case 0:
					lstrcpy(szSend,"Hearts"); break;
				case 1:
					lstrcpy(szSend,"Diamonds");	break;
				case 2:
					lstrcpy(szSend,"Clubs"); break;
				case 3:
					lstrcpy(szSend,"Spades"); break;
				default:
					lstrcpy(szSend,"");	break;
			}
			ExtTextOut(hDC,10,HANDY+20,NULL,NULL,szSend,strlen(szSend),(LPINT)NULL);
		}

		// Paint our hand at the bottom
		UINT nCards = MyHand->GetNumCards();
		if (nCards < nPrevNumCardsHand)
			Rectangle(hDC,HANDX-1,HANDY+29,(nPrevNumCardsHand*CARD_OVERLAP)+HANDX+17,HANDY+55);
		for (i=0; i<nCards; i++)
			PaintCard(hDC,MyHand->PeekCard(i),HANDX+(i*CARD_OVERLAP),HANDY+30);

		// Paint our palette at the bottom
		nCards = MyPal->GetNumCards();
		if (nCards < nPrevNumCardsPal)
			Rectangle(hDC,HANDX-1,HANDY-1,(nPrevNumCardsPal*CARD_OVERLAP)+HANDX+17,HANDY+25);
		for (i=0; i<nCards; i++)
			PaintCard(hDC,MyPal->PeekCard(i),HANDX+(i*CARD_OVERLAP),HANDY);

		// Paint direction arrow
		HBITMAP hBMP = LoadBitmap(hInst,(LPCTSTR)400+bDirFwds);
		HDC hBMPDC = CreateCompatibleDC(hDC);
		HGDIOBJ hSaveObj = SelectObject(hBMPDC,hBMP);
		BitBlt(hDC,rWin.right-150,65,135,53,hBMPDC,0,0,SRCCOPY);
		SelectObject(hBMPDC,hSaveObj);
		DeleteObject(hBMP);
		DeleteDC(hBMPDC);
	}
	else
	{
		Rectangle(hDC,0,0,rWin.right,rWin.bottom);
	}

	EndPaint(hWnd,&ps);							//--PAINT END---

	nPrevNumCardsHand = MyHand->GetNumCards();
	nPrevNumCardsPal = MyPal->GetNumCards();
}








//============================================================ MAIN WINDOW PROC ==######################
LRESULT CALLBACK MainWndProc(HWND hWnd,UINT message,WPARAM wParam,LPARAM lParam)
{
	static HDC	hDC;
//	UINT i;
	CARD cAddCard,cTopPal;

	switch(message) // wedger.demon.co.uk
	{
		case WM_CREATE: 
			break;

		case WSA_READ: // Any async socket event ---
			DoSocketMsg((SOCKET)wParam,WSAGETSELECTEVENT(lParam));
			break;

		case WM_COMMAND:
			switch (LOWORD(wParam))
			{
				case IDM_CONNECT:
					DialogBox(hInst,(LPSTR)IDD_CONNECT,hMainWin,(DLGPROC)ConnectDlgProc);
					break;

				case IDM_DISCONNECT:
					closesocket(sock);
					SetWindowText(hMainWin,szAppName);
					SetWindowText(hChatDlg,"Chat window");
					EnableMenuItem(GetMenu(hMainWin),IDM_CONNECT,MF_ENABLED);
					EnableMenuItem(GetMenu(hMainWin),IDM_DISCONNECT,MF_GRAYED);
					bGameRunning=FALSE;
					break;

				case IDM_ABOUT:
					DialogBox(hInst,(LPSTR)IDD_ABOUT,hMainWin,(DLGPROC)AboutDlgProc);
					break;

				case IDM_PLAY:
					EnableMenuItem(GetMenu(hMainWin),IDM_WATCH,MF_ENABLED);
					EnableMenuItem(GetMenu(hMainWin),IDM_PLAY,MF_GRAYED);
					break;

				case IDM_WATCH:
					EnableMenuItem(GetMenu(hMainWin),IDM_WATCH,MF_GRAYED);
					EnableMenuItem(GetMenu(hMainWin),IDM_PLAY,MF_ENABLED);
					break;

				case IDM_EXIT:
					DestroyWindow(hMainWin);
					break;
			}
			break;

		case WM_PAINT:
			if (!IsIconic(hWnd))
				PaintMainWindow(hWnd);
			break;

			case WM_RBUTTONDOWN:
				MyHand->Order();
				InvalidateRect(hMainWin,NULL,FALSE);
				break;

		case WM_LBUTTONDOWN:
			switch (CheckMouseOver(LOWORD(lParam),HIWORD(lParam)))
			{
				case OVER_PICKUP:
					if (bOurTurn)
					{
						lstrcpy(szSend,"WNUCU$");
						send(sock,szSend,strlen(szSend),NO_FLAGS_SET);
					}
					else
						MessageBeep(MB_ICONHAND);
					break;

				case OVER_PUTDOWN:
					if (bOurTurn)
					{
						CARD cBotPal = MyPal->PeekCard(0);
						UINT nOK=0;
						if (nPickup)
						{
							if (CNUM(cTopCard) == 2) {nOK=0; if (CNUM(cBotPal) == 2) nOK++; }	// 2s
							if ( (CNUM(cTopCard) == 11) && (CSUIT(cTopCard) > 1) )		// Black Jacks
								{nOK=0; if (CNUM(cBotPal) == 11) nOK++; }
						}
						else
						{
							if (CSUIT(cBotPal) == CSUIT(cTopCard)) nOK++;				// Same suit
							if (CNUM(cBotPal) == CNUM(cTopCard)) nOK++;					// Same number
							if (CNUM(cTopCard) == 1)									// Selected ace
								{ nOK=0; if (CSUIT(cBotPal) == nAceSuit) nOK++; }
							if (CNUM(cBotPal) == 1)  nOK++;								// Ace
						}
						if (nOK)
						{
							sprintf(szSend,"WNUCD%u$",MyPal->GetNumCards());
							CARD cTemp;
							char szTemp[10];
							while (MyPal->GetNumCards())
							{
								cTemp = MyPal->TakeCard(0);
								sprintf(szTemp,"%u$",cTemp);
								lstrcat(szSend,szTemp);
							}
							if (CNUM(cTemp)==1)	// Its an ace, choose suit
							{
								nAceSuit = DialogBox(hInst,(LPSTR)IDD_SUIT,hMainWin,(DLGPROC)AboutDlgProc);
								sprintf(szTemp,"%u$",nAceSuit);
								lstrcat(szSend,szTemp);
							}
							send(sock,szSend,strlen(szSend),NO_FLAGS_SET);
							InvalidateRect(hMainWin,NULL,FALSE);
						}
						else
							MessageBeep(MB_ICONHAND);
					}
					else
						MessageBeep(MB_ICONHAND);
					break;

				case OVER_HAND:
					cAddCard = MyHand->PeekCard(nMouseOverCard);
					cTopPal = MyPal->PeekTopCard();
					if (!MyPal->GetNumCards())
					{
						MyPal->GiveCard(MyHand->TakeCard(nMouseOverCard));
						InvalidateRect(hMainWin,NULL,FALSE);
						break;
					}
					if ( (CSUIT(cAddCard) == CSUIT(cTopPal)) || ((CNUM(cTopPal) == 1) || (CNUM(cAddCard) == 1)) )
					{
						int diff = CNUM(cAddCard) - CNUM(cTopPal);
						if ( ((diff>-2) && (diff<2)) || ((diff==12) || (diff==-12)) )
						{
							MyPal->GiveCard(MyHand->TakeCard(nMouseOverCard));
							InvalidateRect(hMainWin,NULL,FALSE);
							break;
						}
					}
					if (CNUM(cAddCard) == CNUM(cTopPal))
					{
						MyPal->GiveCard(MyHand->TakeCard(nMouseOverCard));
						InvalidateRect(hMainWin,NULL,FALSE);
						break;
					}
					MessageBeep(MB_ICONHAND);
					break;

				case OVER_PALETTE:
					while (MyPal->GetNumCards() > nMouseOverCard)
						MyHand->GiveCard(MyPal->TakeCard(nMouseOverCard));
					InvalidateRect(hMainWin,NULL,FALSE);
					break;
			}
			// Fall thru

		case WM_MOUSEMOVE:
			if (CheckMouseOver(LOWORD(lParam),HIWORD(lParam)) == OVER_NOTHING)
				SetCursor(LoadCursor(NULL,IDC_ARROW));
			else
				SetCursor(LoadCursor(hInst,(LPCSTR)IDC_HAND_PRESS));
			return 0;

		case WM_CLOSE:
			DestroyWindow(hMainWin);
			break;

		case WM_DESTROY:
			WSACleanup();
			delete MyHand;
			DeleteObject(hBkgBrush);
			DeleteObject(hYellowPen);
			DeleteObject(hBlackPen);
			PostQuitMessage(0);
			break;
	}
	return(DefWindowProc(hWnd,message,wParam,lParam));
}







BOOL CALLBACK EnumWinUNOProc(HWND hWnd,LPARAM)
{
	GetClassName(hWnd,szBuff,sizeof(szBuff));
	if (!lstrcmp(szMainClassName,szBuff)) nInstances++;
	return TRUE;
}



//===========================================================================================
BOOL ProgramInit(void)
{
	InitCommonControls();

	EnumWindows((WNDENUMPROC)EnumWinUNOProc,NULL);
	UINT nOff = nInstances * 20;
 
	hBkgBrush = CreateSolidBrush(RGB(0,128,0));
	hYellowPen = CreatePen(PS_SOLID,0,RGB(255,255,0));
	hBlackPen = CreatePen(PS_SOLID,0,RGB(0,0,0));

	WNDCLASS wc;
	wc.style = CS_VREDRAW | CS_HREDRAW | CS_BYTEALIGNCLIENT;
	wc.lpfnWndProc = MainWndProc;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = hInst;
	wc.hIcon = LoadIcon(hInst,(LPSTR)IDI_APP);
	wc.hCursor = LoadCursor(NULL,IDC_ARROW);
	wc.hbrBackground = hBkgBrush;
	wc.lpszMenuName = (LPSTR)IDR_MENU;
	wc.lpszClassName = szMainClassName;
	if (!RegisterClass(&wc)) return FALSE;

	hMainWin = CreateWindow(szMainClassName,szAppName,
		WS_OVERLAPPEDWINDOW,40+nOff,20+nOff,420,220,NULL,NULL,hInst,NULL);

	if (!hMainWin)
	{
		MessageBox(NULL,"Cannot create window","Startup error",MB_OK | MB_APPLMODAL);
		return FALSE;
	}
	
	hChatDlg = CreateDialog(hInst,(LPSTR)IDD_CHAT,hMainWin,(DLGPROC)ChatDlgProc);
	SetWindowPos(hChatDlg,NULL,40+nOff,250+nOff,0,0,SWP_NOACTIVATE | SWP_NOSIZE | SWP_SHOWWINDOW);

	hDealerDlg = CreateDialog(hInst,(LPSTR)IDD_DEALER,hMainWin,(DLGPROC)DealerDlgProc);
	SetWindowPos(hDealerDlg,NULL,470+nOff,20+nOff,0,0,SWP_NOACTIVATE | SWP_NOSIZE);


	// ===================================== init Winsock ====

	WSADATA WSAData;
	char szTemp[80];

	if ((status = WSAStartup(MAKEWORD(1,1),&WSAData)) != 0) 
	{
		sprintf(szTemp,"%d is the err",status);
		MessageBox(hMainWin,szTemp,"Error",MB_OK);
	}


	// ===================================== init game ====

	MyHand = new HAND(FALSE);
	MyPal = new HAND(FALSE);

	return TRUE;
}




// ### Entry point
//===========================================================================================
int PASCAL WinMain(HINSTANCE hInstance,HINSTANCE,LPSTR,int)
{
	MSG msg;
	hInst = hInstance; // Squirrel our instance ID

	if (!ProgramInit()) return -1;

	ShowWindow(hMainWin,SW_SHOWNORMAL);
	UpdateWindow(hMainWin);

	// ======================== MESSAGE LOOP ========================

	while (GetMessage(&msg,NULL,0,0))
	{
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}
	return msg.wParam;
}
