// server.cpp -- Wedges Internet UNO Server
//
// By Andrew Wedgbury
// Copyright 1997

#define STRICT
#include <windows.h>
#include <winsock.h>
#include <windowsx.h>
#include <commctrl.h>
#include <string.h>
#include <stdio.h>
#include <time.h>
#include <math.h>
#pragma hdrstop

#include "resource.h"
#include "..\NetUNO.h"

//====================================================== GLOBAL DATA ====
const char	szAppName[] = "NetUNO Server";
const char	szWelcome[] = "--- Welcome to Wedge's NetUNO Server ---";
const char	szVersion[] = "1.00 (32bit)";

#define MAX_PENDING_CONNECTS	4 
#define NO_FLAGS_SET			0
#define UNO_MSG_LENGTH			400
#define WSA_ACCEPT			(WM_USER + 0)
#define WSA_READ			(WM_USER + 1)

HINSTANCE	hInst;						// Windows data
HWND		hMainWin,hListView,hStatus;
UINT		nPingTimer;
DWORD		dwStartTime=0;

SOCKET		sock;						// Winsock data
u_short		portno = 1527;
int			status;
SOCKADDR_IN local_sin,acc_sin;
int			acc_sin_len;

typedef struct _UNOCLIENT
{
	BOOL		bConnected;
	BOOL		bIn;
	char		szName[20];
	UINT		nCards;
	SOCKET		socket;
	SOCKADDR_IN	sin;
} UNOCLIENT;

char		szBuff[UNO_MSG_LENGTH];
char		szSend[UNO_MSG_LENGTH+30];
char		szDispText[200];

UINT		nClients=0;					// Game data
UNOCLIENT	Clients[10];
UINT		nPlayersIn=0;
BOOL		bDirFwds=TRUE;
BOOL		bRuns=TRUE;
UINT		nTurn=0;
UINT		nDealer=0;
UINT		nDecks=1;
UINT		nCardsEach=7;
UINT		nPickup=0;
UINT		nAceSuit=0;

HAND		*Deck;
HAND		*Pile;

//======================== Function Prototypes ==
int PASCAL			WinMain(HINSTANCE,HINSTANCE,LPSTR,int);
LRESULT CALLBACK	MainWndProc(HWND,UINT,WPARAM,LPARAM);	// WNDPROC's
BOOL				ProgramInit(void);


//========================================== Broadcast CHAT text =====
void BroadcastChatText(LPCSTR szText,char szName[]=NULL)
{
	char szBc[UNO_MSG_LENGTH+30];
	if (szName == NULL)
		sprintf(szBc,"WNUST%u$%s",lstrlen(szText),szText);
	else
		sprintf(szBc,"WNUST%u$%s: %s",lstrlen(szText)+lstrlen(szName)+2,szName,szText);
	
	for (UINT c=0; c<nClients; c++)
	{
		send(Clients[c].socket,szBc,strlen(szBc),NO_FLAGS_SET);
	}
}


void SendChatText(LPCSTR szText,UINT c)
{
	char szBc[UNO_MSG_LENGTH+30];
	sprintf(szBc,"WNUST%u$%s",lstrlen(szText),szText);
	send(Clients[c].socket,szBc,strlen(szBc),NO_FLAGS_SET);
}



void Turn(void)
{
	UINT i=0;
	do
	{
		if (bDirFwds)
			{ nTurn++; if (nTurn >= nClients) nTurn=0; }
		else
			{ if (nTurn == 0) nTurn = nClients;	nTurn--; }
		if (++i >= nClients) break; // watchdog
	} while (!Clients[nTurn].bIn);
}


void NextTurn(UINT n8s=0,UINT n10s=0)
{
	UINT i;
	if (n10s)
	{
		for (i=0; i<n10s; i++)
		{
			bDirFwds = (!bDirFwds);
			Turn();
		}
		return;
	}
	if (n8s)
	{
		for (i=0; i<=n8s; i++)
		{
			Turn();
		}
		return;
	}
	Turn();
}


void BroadcastGameStats(void)
{
	char szBc[UNO_MSG_LENGTH+30];
	char szTemp[80],szPrint[10];
	UINT c;
	lstrcpy(szTemp,"");

	sprintf(szTemp,"%u$",Pile->GetNumCards());
	for (c=0; c<(Pile->GetNumCards()); c++)
	{
		sprintf(szPrint,"%u$",(UINT)Pile->PeekCard(c));
		lstrcat(szTemp,szPrint);
	}

	for (c=0; c<nClients; c++)
	{
		sprintf(szPrint,"%u$",(UINT)Clients[c].nCards);
		lstrcat(szTemp,szPrint);
	}

	for (c=0; c<nClients; c++)
	{
		
		sprintf(szBc,"WNUSN%u$%u$%u$%u$%u$%s",(UINT)bDirFwds,nTurn,(UINT)(nTurn==c),nPickup,nAceSuit,szTemp); // Pile
		send(Clients[c].socket,szBc,lstrlen(szBc),NO_FLAGS_SET);
	}
	
}


void BroadcastGameData(void)
{
	char szBc[UNO_MSG_LENGTH+30];
	sprintf(szBc,"WNUSD%u$",nClients);

	UINT c;
	for (c=0; c<nClients; c++)
	{
		lstrcat(szBc,Clients[c].szName);
		lstrcat(szBc,"$");
	}
	
	for (c=0; c<nClients; c++)
	{
		send(Clients[c].socket,szBc,lstrlen(szBc),NO_FLAGS_SET);	
	}
}


void SendWelcome(UINT c)
{
	SendChatText("",c);
	SendChatText(szWelcome,c);
	SendChatText("Players currently online are:",c);
	
	// Send a list of all the names
	lstrcpy(szSend,"");
	for (UINT d=0; d<nClients; d++)
	{
		if (d) lstrcat(szSend,", ");
		lstrcat(szSend,Clients[d].szName);
	}
	SendChatText(szSend,c);
	SendChatText("",c);
}


UINT Parse(char szSrc[],char *szDest)
{
	UINT i;
	for (i=0; i<200; i++)
	{
		if (szSrc[i] == '$') break;
		*(szDest + i) = szSrc[i];	
	}
	*(szDest + i) = 0;
	return i+1;
}



BOOL StringIs(LPCTSTR szStr1,LPCTSTR szStr2)
{
	UINT len = lstrlen(szStr2);
	if (CompareString(LOCALE_SYSTEM_DEFAULT,NORM_IGNORECASE,szStr1,len,szStr2,len) ==2)
		return TRUE;
	else
		return FALSE;
}



void DoEscapeCommand(char szMsg[],UINT c)
{
	char szTemp[400];
	if (StringIs(szMsg,"/STATS")) // Server statistics
	{
		sprintf(szTemp,"--- Wedge's NetUNO Server version %s (1997 Wedge) ---",szVersion); SendChatText(szTemp,c);
		gethostname(szBuff,sizeof(szBuff));
		sprintf(szTemp,"Running on: %s",szBuff); SendChatText(szTemp,c);
		char szTime[40];
		GetTimeFormat(NULL,NULL,NULL,NULL,szTime,sizeof(szTime));
		sprintf(szTemp,"Local time: %s",szTime); SendChatText(szTemp,c);
		DWORD dwTime = (GetTickCount() - dwStartTime);
		int Hours = dwTime / 3600000;
		int Minutes = (dwTime - (Hours * 3600000)) / 60000;
		int Seconds = (dwTime - (Hours * 3600000) - (Minutes * 60000)) / 1000;
		sprintf(szTemp,"Server uptime: %.2u:%.2u:%.2u",Hours,Minutes,Seconds); SendChatText(szTemp,c);
		sprintf(szTemp,"Number of sockets: %u",nClients); SendChatText(szTemp,c);
		MEMORYSTATUS memstat; memstat.dwLength = sizeof(memstat);
		GlobalMemoryStatus(&memstat);
		sprintf(szTemp,"Free memory: %uKB (%u%%)",((memstat.dwAvailPhys+memstat.dwAvailPageFile)/1000),memstat.dwMemoryLoad); 
		SendChatText(szTemp,c);
	}
	if (StringIs(szMsg,"/SHOW")) // Server show window
		ShowWindow(hMainWin,SW_SHOW);
	if (StringIs(szMsg,"/HIDE")) // Server hide window
		ShowWindow(hMainWin,SW_HIDE);
	if (StringIs(szMsg,"/SHUTDOWN")) // Server shutdown
		DestroyWindow(hMainWin);


	sprintf(szTemp,"-OK-"); SendChatText(szTemp,c);
}




void DisconnectPlayer(UINT c)
{
	ListView_DeleteItem(hListView,c);
	sprintf(szSend,"--- %s has quit NetUNO, bye %s ---",Clients[c].szName,Clients[c].szName);
	Clients[c].bConnected = FALSE;
	if (Clients[c].bIn) nPlayersIn--;
	BroadcastChatText(szSend);
	
	nClients--;
	if (nClients)
		for (UINT i=c; i<nClients; i++)
			Clients[i] = Clients[i+1];

	BroadcastGameData();
	sprintf(szSend,"%u Connections",nClients);
	SetWindowText(hStatus,szSend);
}





void ExecuteSocketMsg(char szMsg[],UINT c)
{
	UINT d,n,x;
	UINT n8s,n10s;
	char szParse[200];
	CARD cTemp;
	BOOL bResetPickup;

	switch (szMsg[4])				// CLIENT -> SERVER Messages
	{
		case 'C': // CONNECT
			// Get name and update list view
			Clients[c].bConnected = TRUE;
			lstrcpy(Clients[c].szName,&szMsg[5]);
			ListView_SetItemText(hListView,c,0,Clients[c].szName);

			// Send confirmation back
			sprintf(szSend,"WNUSC%u$%u$",c,(UINT)nPlayersIn);
			send(Clients[c].socket,szSend,strlen(szSend),NO_FLAGS_SET);

			// Tell everyone who has joined
			sprintf(szSend,"--- %s has joined NetUNO ---",Clients[c].szName);
			BroadcastChatText(szSend);

			if (c == 0)
			{
				sprintf(szSend,"WNUSS1$");
				send(Clients[c].socket,szSend,strlen(szSend),NO_FLAGS_SET);
			}

			SendWelcome(c);
			BroadcastGameData();
			BroadcastGameStats();
			break;

		case 'T': // TEXT CHAT
			n = 5;
			n += Parse(&szMsg[n],szParse);
			memcpy(szSend,&szMsg[n],atoi(szParse)); szSend[atoi(szParse)] =0;
			if (szSend[0] == '/')			// escape seq command
				DoEscapeCommand(szSend,c);
			else							// regular chat txt
				BroadcastChatText(szSend,Clients[c].szName);
			break;

		case 'D': // PUT DOWN
			n = 5;
			n += Parse(&szMsg[n],szParse); d = atoi(szParse);
			for (x=0; x<d; x++)
			{
				n += Parse(&szMsg[n],szParse);
				cTemp = atoi(szParse);
				if (cTemp !=255);
				{
					Pile->GiveCard(cTemp);
					Clients[c].nCards--;
				}
			}
			bResetPickup=TRUE; n8s=0; n10s=0;

			if (CNUM(cTemp)==1)								// Get Ace suit
			{
				n += Parse(&szMsg[n],szParse); nAceSuit=atoi(szParse);
			}

			if (CNUM(cTemp)==2)								// Trace back 2s
			{
				for(x=(Pile->GetNumCards()-1); x>=(Pile->GetNumCards()-d); x--)
				{
					if (CNUM(Pile->PeekCard(x)) == 2)
						{ nPickup +=2; bResetPickup=FALSE; }
					else break;
				}
			}

			if ( (CNUM(cTemp)==11) && (CSUIT(cTemp)>1))		// Trace back Black Jacks
			{
				for(x=(Pile->GetNumCards()-1); x>=(Pile->GetNumCards()-d); x--)
				{
					if ( (CNUM(Pile->PeekCard(x)) == 11) && (CSUIT(Pile->PeekCard(x))>1) )
						{ nPickup +=5; bResetPickup=FALSE; }
					else break;
				}
			}

			if (CNUM(cTemp)==8)								// Trace back 8s
			{
				for(x=(Pile->GetNumCards()-1); x>=0; x--)
				{
					if (CNUM(Pile->PeekCard(x)) == 8)
						{ n8s++; if (n8s=d) break; }
					else break;
				}
			}

			if (CNUM(cTemp)==10)							// Trace back 10s
			{
				for(x=(Pile->GetNumCards()-1); x>=0; x--)
				{
					if (CNUM(Pile->PeekCard(x)) == 10)
						{ n10s++; if (n10s=d) break; }
					else break;
				}
			}

			if (bResetPickup) nPickup=0;

			while (Pile->GetNumCards() > 5)
				Deck->GiveCardBottom(Pile->TakeCard(0));

			if (Clients[c].nCards == 0)						// Player is out
			{
				// Tell everyone who is out
				sprintf(szSend,"--- %s is out ---",Clients[c].szName);
				BroadcastChatText(szSend);
				Clients[c].bIn=FALSE;
				if ((--nPlayersIn)<=1)
				{
					// Tell everyone who is out
					lstrcpy(szSend,"--- GAME OVER---");
					BroadcastChatText(szSend);
					break;
				}
			}
			NextTurn(n8s,n10s);
			BroadcastGameStats();
			break;

		case 'U': // PICK UP
			if (!nPickup) nPickup=1;
			for (x=0; x<nPickup; x++)
			{
				Clients[c].nCards++;
				sprintf(szSend,"WNUSU%u$",Deck->TakeTopCard());
				send(Clients[c].socket,szSend,strlen(szSend),NO_FLAGS_SET);
			}
			nPickup=0;
			NextTurn();
			BroadcastGameStats();
			break;

		case 'P': // PLAYER SETS PLAY/WATCH MODE
			n = 5;
			n += Parse(&szMsg[n],szParse);	x = atoi(szParse);
/*			if (x)
			{
							Clients[c].bIn
			}
*/			sprintf(szSend,"--- %s is not playing ---",Clients[c].szName);
			BroadcastChatText(szSend);
			break;

		case 'S': // DEALER SETS NEW DEALER
			n = 5;
			n += Parse(&szMsg[n],szParse);
			x = atoi(szParse);
			// Send new dealer info out
			for (d=0; d<nClients; d++)
			{
				if (d == x)	sprintf(szSend,"WNUSS1$");
				else sprintf(szSend,"WNUSS0$");
				send(Clients[d].socket,szSend,strlen(szSend),NO_FLAGS_SET);
			}
			sprintf(szSend,"--- %s is now the dealer ---",Clients[x].szName);
			BroadcastChatText(szSend);
			break;

		case 'G': // DEALER REQUESTED NEW GAME
			n = 5;
			n += Parse(&szMsg[n],szParse);	nDecks = atoi(szParse);
			n += Parse(&szMsg[n],szParse); nCardsEach = atoi(szParse);
			n += Parse(&szMsg[n],szParse); nTurn = atoi(szParse);
			n += Parse(&szMsg[n],szParse); bDirFwds = atoi(szParse);
			n += Parse(&szMsg[n],szParse); bRuns = atoi(szParse);

			nPlayersIn = nClients; // All players in

			// Announce new game
			sprintf(szSend,"--- NEW GAME - %u Players ---",nPlayersIn);
			BroadcastChatText(szSend);

			delete Deck; Deck = new HAND(TRUE);
			delete Pile; Pile = new HAND(FALSE);
			Deck->Shuffle();
			nPickup=0;

			sprintf(szSend,"WNUSG");
			for (d=0; d<nClients; d++)
			{
				Clients[d].bIn=TRUE;
				Clients[d].nCards=0;
				send(Clients[d].socket,szSend,strlen(szSend),NO_FLAGS_SET);
			}

			// Deal cards

			for (x=0; x<nCardsEach; x++)
			{
				for (d=0; d<nClients; d++)
				{
					Clients[d].nCards++;
					sprintf(szSend,"WNUSU%u$",Deck->TakeTopCard());
					send(Clients[d].socket,szSend,strlen(szSend),NO_FLAGS_SET);
				}
			}
			Pile->GiveCard(Deck->TakeTopCard());
			BroadcastGameStats();
			break;

		case 'Q': // QUIT
			break;
	}
}







//========================================== Handle socket messages =====
void DoSocketMsg(SOCKET s,WORD msg)
{
	UINT i,c;
	// Find client number from socket
	for (c=0; c<=nClients; c++)
	{
		if (c == nClients) return;
		if (Clients[c].socket == s) break;
	}
	// we are dealing with client c

	if (msg == FD_READ) // Socket READ
	{
		status = recv(s,szBuff,UNO_MSG_LENGTH,NO_FLAGS_SET);
		if (status) 
		{
			szBuff[status] = '\0';
			for (i=0; i<(UINT)status; i++)
			{
				if (( (szBuff[i] == 'W') && (szBuff[i+1] == 'N') ) && ( (szBuff[i+2] == 'U') && (szBuff[i+3] == 'C') ))
				{
					ExecuteSocketMsg(&szBuff[i],c);
				}
			}
			return;
		}
	}
	DisconnectPlayer(c);
}


//============================================================ MAIN WINDOW PROC ==######################
LRESULT CALLBACK MainWndProc(HWND hWnd,UINT message,WPARAM wParam,LPARAM lParam)
{
	static HDC	hDC;

	switch(message) // wedger.demon.co.uk
	{
		case WM_CREATE: 
//			nPingTimer = SetTimer(hMainWin,100,5000,(TIMERPROC)AdvertsTimerProc);
			break;

		case WSA_ACCEPT:
			if (WSAGETSELECTERROR(lParam) == 0) 
			{
				acc_sin_len = sizeof(acc_sin);
				Clients[nClients].socket = accept(sock,(struct sockaddr FAR *)&acc_sin,(int FAR *)&acc_sin_len);
				Clients[nClients].sin = acc_sin;
				Clients[nClients].bConnected = FALSE;

				LV_ITEM lvitem;						// Add name to list view
				lvitem.mask = LVIF_TEXT | LVIF_IMAGE;
				lvitem.iItem = nClients; lvitem.iSubItem = 0;
				lvitem.pszText = "...";
				lvitem.cchTextMax = 20;
				lvitem.iImage = 0;
				ListView_InsertItem(hListView,&lvitem);

				sprintf(szBuff,"%u.%u.%u.%u",
					acc_sin.sin_addr.S_un.S_un_b.s_b1,
					acc_sin.sin_addr.S_un.S_un_b.s_b2,
					acc_sin.sin_addr.S_un.S_un_b.s_b3,
					acc_sin.sin_addr.S_un.S_un_b.s_b4);

				lvitem.mask = LVIF_TEXT;			// Add version to list view
				lvitem.iItem = nClients; lvitem.iSubItem = 1;
				lvitem.pszText = szBuff;
				lvitem.cchTextMax = 30;
				ListView_SetItem(hListView,&lvitem);
				
				if ((status = WSAAsyncSelect(Clients[nClients].socket,hWnd,WSA_READ,FD_READ | FD_CLOSE )) > 0) 
					closesocket(Clients[nClients].socket);
				
				Clients[nClients].nCards=0;
				Clients[nClients].bIn=FALSE;

				nClients++; // Confirm connection
				sprintf(szSend,"%u Connections",nClients);
				SetWindowText(hStatus,szSend);
			}
			break;

		case WSA_READ: // Any async socket event ---
			DoSocketMsg((SOCKET)wParam,WSAGETSELECTEVENT(lParam));
			break;

		case WM_SYSCOMMAND:
			if (wParam == SC_MINIMIZE)
			{
				ShowWindow(hMainWin,SW_HIDE);
				return 0;
			}
			break;

		case WM_CLOSE:
			DestroyWindow(hMainWin);
			break;

		case WM_DESTROY:
			KillTimer(hWnd,nPingTimer);
			WSACleanup();
			PostQuitMessage(0);
			break;
	}
	return(DefWindowProc(hWnd,message,wParam,lParam));
}





//===========================================================================================
BOOL ProgramInit(void)
{
	const char	szMainClassName[] = "WUNOServ";

	InitCommonControls();
	srand((unsigned)time(NULL));

	dwStartTime = GetTickCount();

	WNDCLASS wc;
	wc.style = CS_VREDRAW | CS_HREDRAW | CS_BYTEALIGNCLIENT;
	wc.lpfnWndProc = MainWndProc;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = hInst;
	wc.hIcon = LoadIcon(hInst,(LPSTR)IDI_APP);
	wc.hCursor = LoadCursor(NULL,IDC_ARROW);
	wc.hbrBackground = (HBRUSH)GetStockObject(LTGRAY_BRUSH);
	wc.lpszMenuName = (LPSTR)NULL;
	wc.lpszClassName = szMainClassName;
	if (!RegisterClass(&wc)) return FALSE;

	hMainWin = CreateWindow(szMainClassName,szAppName,
		WS_OVERLAPPED | WS_CAPTION | WS_SYSMENU | WS_MINIMIZEBOX,
		GetSystemMetrics(SM_CXSCREEN)-220,20,
		200,180,NULL,NULL,hInst,NULL);

	if (!hMainWin)
	{
		MessageBox(NULL,"Cannot create window","Startup error",MB_OK | MB_APPLMODAL);
		return FALSE;
	}
	
	hListView = CreateWindow(WC_LISTVIEW,"",WS_CHILD | WS_VISIBLE | LVS_REPORT,0,0,195,138,hMainWin,NULL,hInst,NULL);

	LV_COLUMN	col;								// Set up image list
	col.mask = LVCF_FMT | LVCF_WIDTH | LVCF_TEXT;
	col.fmt = LVCFMT_LEFT; col.cx = 93; col.pszText = "Name"; col.cchTextMax = 20;
	ListView_InsertColumn(hListView,1,&col);
		
	col.mask = LVCF_FMT | LVCF_WIDTH | LVCF_TEXT;
	col.fmt = LVCFMT_LEFT; col.cx = 100; col.pszText = "IP address"; col.cchTextMax = 20;
	ListView_InsertColumn(hListView,2,&col);

    HIMAGELIST himlSmall = ImageList_Create(GetSystemMetrics(SM_CXSMICON), 
        GetSystemMetrics(SM_CYSMICON),TRUE,1,1); 
    HICON hiconItem = LoadIcon(hInst, MAKEINTRESOURCE(IDI_CLIENT)); 
    ImageList_AddIcon(himlSmall,hiconItem); 
    DeleteObject(hiconItem); 
    ListView_SetImageList(hListView, himlSmall, LVSIL_SMALL); 

	hStatus = CreateWindow("STATIC","Status",WS_CHILD | WS_VISIBLE | SS_LEFT,2,138,197,18,hMainWin,NULL,hInst,NULL);


	// ===================================== init Winsock ====

	WSADATA WSAData;
	char szTemp[80];

	if ((status = WSAStartup(MAKEWORD(1,1),&WSAData)) != 0) 
	{
		sprintf(szTemp,"%d is the err",status);
		MessageBox(hMainWin,szTemp,"Error",MB_OK);
	}

//	gethostname(szBuff,sizeof(szBuff));
//	sprintf(szSend,"UNOServer - %s",szBuff);
//	SetWindowText(hMainWin,szSend);

	sock = socket( AF_INET, SOCK_STREAM, 0);
	if (sock == INVALID_SOCKET) 
	{
		MessageBox(hMainWin,"socket() failed","Error",MB_OK);
		return FALSE;
	}

	local_sin.sin_family = AF_INET;
	DWORD dwSize = sizeof(szBuff);
	gethostname(szBuff,dwSize);
	local_sin.sin_addr.s_addr = INADDR_ANY;
	local_sin.sin_port = htons(portno);

	if (bind(sock,(struct sockaddr FAR *)&local_sin,sizeof(local_sin)) == SOCKET_ERROR) 
	{
		sprintf(szBuff,"%d is the error",WSAGetLastError());
		MessageBox(hMainWin,szBuff,"bind(sock) failed",MB_OK);
		closesocket(sock);
        return FALSE;
	}

	if (listen(sock,MAX_PENDING_CONNECTS) < 0)
	{
		sprintf(szBuff,"%d is the error",WSAGetLastError());
		MessageBox(hMainWin,szBuff,"listen(sock) failed",MB_OK);
		return FALSE;
	}
	
	if ((status = WSAAsyncSelect(sock,hMainWin,WSA_ACCEPT,FD_ACCEPT)) > 0) 
	{
		wsprintf(szBuff,"%d (0x%x)");
		MessageBox(hMainWin,"Error on WSAAsyncSelect()",szBuff,MB_OK);
		closesocket(sock);
	}


	// ===================================== init game ====

	Deck = new HAND(TRUE);
	Pile = new HAND(FALSE);

	sprintf(szSend,"%u Connections",nClients);
	SetWindowText(hStatus,szSend);

	return TRUE;
}




// ### Entry point
//===========================================================================================
int PASCAL WinMain(HINSTANCE hInstance,HINSTANCE,LPSTR,int)
{
	MSG msg;
	hInst = hInstance; // Squirrel our instance ID

	if (!ProgramInit()) return -1;

	ShowWindow(hMainWin,SW_SHOWNORMAL);
	UpdateWindow(hMainWin);

	// ======================== MESSAGE LOOP ========================

	while (GetMessage(&msg,NULL,0,0))
	{
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}
	return msg.wParam;
}
